#if !defined  HAVE_BITPOLMODMULT_H__
#define       HAVE_BITPOLMODMULT_H__

#include "fxttypes.h"


static inline ulong bitpolmod_times_x(ulong a, ulong c, ulong h)
// Return  (A * x) mod C
// where A and C represent polynomials over Z/2Z:
//  W = pol(w) =: \sum_k{ [bit_k(w)] * x^k}
//
// h needs to be a mask with one bit set:
//  h == highest_bit(c) >> 1  == 1UL << (degree(C)-1)
//
// If C is a primitive polynomial of degree n
// successive calls will cycle through all 2**n-1
// n-bit words and the sequence of bits
// (any fixed position) of a constitutes
// a shift register sequence (SRS).
// Start with a=2 to get a SRS that starts with
// n-1 consecutive zeroes (use bit 0 of a)
{
    ulong s = a & h;
    a <<= 1;
    if ( s )  a ^= c;
    return  a;
}
// -------------------------

static inline ulong bitpolmod_div_x(ulong a, ulong c, ulong h)
// Return  (A / x) mod C
// where A and C represent polynomials over Z/2Z:
//  W = pol(w) =: \sum_k{ [bit_k(w)] * x^k}
// C must have nonzero constant term: (c&1)==1
//
// h needs to be a mask with one bit set:
//  h == highest_bit(c) >> 1  == 1UL << (degree(C)-1)
//
// If C is a primitive polynomial of degree n
// successive calls will cycle through all 2**n-1
// n-bit words and the sequence of bits
// (any fixed position) of a constitutes
// a shift register sequence (SRS).
{
    ulong s = a & 1;
    a >>= 1;
    if ( s )
    {
        a ^= (c>>1);
        a |= h;  // so it also works for  n == BITS_PER_LONG
    }
    return  a;
}
// -------------------------

inline ulong bitpolmod_mult(ulong a, ulong b, ulong c, ulong h)
// Return  (A * B) mod C
// where A, B and C represent polynomials over Z/2Z:
//  W = pol(w) =: \sum_k{ [bit_k(w)] * x^k}
//
// h needs to be a mask with one bit set:
//  h == highest_bit(c) >> 1  == 1UL << (degree(C)-1)
//
// With b=2 (== 'x') the result is identical to
//   bitpolmod_times_x(a, c, h)
{
    ulong t = 0;
    while ( b )
    {
        if ( b & 1 )  t ^= a;
        b >>= 1;

        ulong s = a & h;
        a <<= 1;
        if ( s )  a ^= c;
    }
    return  t;
}
// -------------------------

inline ulong bitpolmod_square(ulong a, ulong c, ulong h)
// Todo: optimize using bit_zip() & table
{
    return bitpolmod_mult(a, a, c, h);
}
// -------------------------


inline ulong bitpolmod_power(ulong a, ulong e, ulong c, ulong h)
// Return (A ** e)  mod C
// where A and C represent polynomials over Z/2Z:
//   W = pol(w) =: \sum_k{ [bit_k(w)] * e^k}
//
// h needs to be a mask with one bit set:
//  h == highest_bit(c) >> 1  == 1UL << (degree(C)-1)
//
// With irreducible C the inverse of A can be obtained via
// i = bitpolmod_power(a, c, r1, h)
//  where r1 = (h<<1)-2 = max_order - 1 = 2^degree(C) - 2
// Then  1 == bitpolmod_mult(a, c, i, h)
{
    if ( 0==a )  return 0;
    if ( 0==e )  return 1;

    ulong s = a;
    while ( 0==(e&1) )
    {
        s = bitpolmod_square(s, c, h);
        e >>= 1;
    }

    a = s;
    while ( 0!=(e>>=1) )
    {
        s = bitpolmod_square(s, c, h);
        if ( e & 1 )  a = bitpolmod_mult(a, s, c, h);
    }
    return  a;
}
// -------------------------



inline ulong bitpolmod_inverse(ulong a, ulong c, ulong h)
// Return (A ** -1)  mod C
// where A and C represent polynomials over Z/2Z:
//   W = pol(w) =: \sum_k{ [bit_k(w)] * e^k}
//
// h needs to be a mask with one bit set:
//  h == highest_bit(c) >> 1  == 1UL << (degree(C)-1)
//
// With irreducible C the inverse of A can be obtained via
// i = bitpolmod_power(a, c, r1, h)
//  where r1 = (h<<1)-2 = max_order - 1 = 2^degree(C) - 2
// Then  1 == bitpolmod_mult(a, c, i, h)
{
    ulong r1 = (h<<1)-2;
    ulong i =  bitpolmod_power(a, c, r1, h);
    return  i;
}
// -------------------------


static inline ulong bitpolmod_sqrt(ulong a, ulong c, ulong h)
// Return sqrt(A)  mod C
// where A and C represent polynomials over Z/2Z:
//   W = pol(w) =: \sum_k{ [bit_k(w)] * e^k}
//
// h needs to be a mask with one bit set:
//  h == highest_bit(c) >> 1  == 1UL << (degree(C)-1)
//
// sqrt(A) = A^(2^(n-1)) where n=deg(C)
//
{
    ulong d = h;
    while ( (d>>=1) )
    {
        a = bitpolmod_square(a, c, h);
    }
    return  a;
}
// -------------------------


#endif  // !defined HAVE_BITPOLMODMULT_H__
